﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Analysis.SolutionCaching.RealVectorEncoding;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("RealVectorConvexHullModifier", "Calculates the convex hull from a solution cache. ")]
  [StorableClass]
  public class RealVectorConvexHullModifier : ParameterizedNamedItem, IRunCollectionModifier {
    private const string SolutionCacheResultName = "SolutionCache";

    [StorableConstructor]
    protected RealVectorConvexHullModifier(bool deserializing) : base(deserializing) { }
    protected RealVectorConvexHullModifier(RealVectorConvexHullModifier original, Cloner cloner) : base(original, cloner) { }
    public RealVectorConvexHullModifier() { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new RealVectorConvexHullModifier(this, cloner);
    }

    public void Modify(List<IRun> runs) {
      foreach (var run in runs) {
        int i = 0;
        if (!run.Results.ContainsKey(SolutionCacheResultName)) continue;
        var solutionCache = run.Results.Single(x => x.Key == "SolutionCache").Value as RealVectorSolutionCache;
        if (solutionCache == null) continue;

        var sols = solutionCache.GetSolutionsFromGeneration(i);
        DataTable volDataTable = new DataTable("Convex hull volume over generations");
        DataTable nrOfPointsTable = new DataTable("Nr. of points of convex hull");
        DoubleValue overallVolumn = new DoubleValue();

        DataRow dtVolumeRow = new DataRow("Volume");
        DataRow dtNrPointsRow = new DataRow("Nr. of points");
        volDataTable.Rows.Add(dtVolumeRow);
        nrOfPointsTable.Rows.Add(dtNrPointsRow);

        List<double[]> curHull = null;

        while (sols.Count != 0) {
          var input = sols.Select(x => ConvertRealVectorToVertexArray(x)).ToArray();
          var convexHull = LPHull.Calculate(input);
          var volume = ConvexHullMeasures.CalculateVolume(convexHull);
          dtVolumeRow.Values.Add(volume);
          dtNrPointsRow.Values.Add(convexHull.Count);

          if (curHull == null) {
            curHull = convexHull;
          } else {
            var newPHull = curHull.Union(convexHull).ToArray();
            curHull = LPHull.Calculate(newPHull);
          }

          sols = solutionCache.GetSolutionsFromGeneration(++i);
        }
        overallVolumn.Value = ConvexHullMeasures.CalculateVolume(curHull);

        run.Results.Add("Overall volumn: ", overallVolumn);
        run.Results.Add("Convex hull volumn", volDataTable);
        run.Results.Add("Convex nr. of points", nrOfPointsTable);
      }
    }

    private double[] ConvertRealVectorToVertexArray(RealVector p) {
      double[] vertex = p.Select(x => (double)x).ToArray();
      return vertex;
    }
  }
}